import os
from typing import Any

import graphene
from graphql import (
    GraphQLArgument,
    GraphQLBoolean,
    GraphQLInt,
    GraphQLNonNull,
    GraphQLString,
)

from graphene_directives import (
    CustomDirective,
    DirectiveLocation,
    Schema,
    SchemaDirective,
    build_schema,
    directive,
)

curr_dir = os.path.dirname(os.path.realpath(__file__))


def input_transform(inputs: dict, _schema: Schema) -> dict:
    """
    def input_transform (inputs: Any, schema: Schema) -> dict,
    """
    if inputs.get("max_age") > 200:
        inputs["swr"] = 30
    return inputs


def validate_non_field_input(_type: Any, inputs: dict, _schema: Schema) -> bool:
    """
    def validator (type_: graphene type, inputs: Any, schema: Schema) -> bool,
    if validator returns False, library raises DirectiveCustomValidationError
    """
    if inputs.get("max_age") > 2500:
        return False
    return True


def validate_field_input(
    _parent_type: Any, _field_type: Any, inputs: dict, _schema: Schema
) -> bool:
    """
    def validator (parent_type_: graphene_type, field_type_: graphene type, inputs: Any, schema: Schema) -> bool,
    if validator returns False, library raises DirectiveCustomValidationError
    """
    if inputs.get("max_age") > 2500:
        return False
    return True


CacheDirective = CustomDirective(
    name="cache",
    locations=[
        DirectiveLocation.OBJECT,
        DirectiveLocation.INTERFACE,
        DirectiveLocation.ENUM,
        DirectiveLocation.UNION,
        DirectiveLocation.INPUT_OBJECT,
        DirectiveLocation.FIELD_DEFINITION,
        DirectiveLocation.INPUT_FIELD_DEFINITION,
        DirectiveLocation.SCALAR,
    ],
    args={
        "max_age": GraphQLArgument(
            GraphQLNonNull(GraphQLInt),
            description="Specifies the maximum age for cache in seconds.",
        ),
        "swr": GraphQLArgument(
            GraphQLInt, description="Stale-while-revalidate value in seconds. Optional."
        ),
        "scope": GraphQLArgument(
            GraphQLString, description="Scope of the cache. Optional."
        ),
    },
    description="Caching directive to control cache behavior of fields or fragments.",
    non_field_validator=validate_non_field_input,
    field_validator=validate_field_input,
    input_transform=input_transform,
)


AuthenticatedDirective = CustomDirective(
    name="authenticated",
    locations=[
        DirectiveLocation.OBJECT,
        DirectiveLocation.INTERFACE,
        DirectiveLocation.ENUM,
        DirectiveLocation.ENUM_VALUE,
        DirectiveLocation.UNION,
        DirectiveLocation.INPUT_OBJECT,
        DirectiveLocation.FIELD_DEFINITION,
        DirectiveLocation.SCALAR,
        DirectiveLocation.ARGUMENT_DEFINITION,
    ],
    args={
        "required": GraphQLArgument(
            GraphQLNonNull(GraphQLBoolean), description="Auth required"
        )
    },
    description="Auth directive to control authorization behavior.",
)

# This will prevent @key definition from being added to the schema, but decorator will work
KeyDirective = CustomDirective(
    name="key",
    locations=[DirectiveLocation.OBJECT, DirectiveLocation.INTERFACE],
    description="Declaring an entity",
    add_definition_to_schema=False,
)

# No argument directive
InternalDirective = CustomDirective(
    name="internal",
    locations=[
        DirectiveLocation.OBJECT,
        DirectiveLocation.FIELD_DEFINITION,
        DirectiveLocation.ARGUMENT_DEFINITION,
    ],
    description="Directive to indicate this is a internal field.",
)

# A Repeatable directive
RepeatableDirective = CustomDirective(
    name="repeatable_directive",
    locations=[DirectiveLocation.OBJECT, DirectiveLocation.FIELD_DEFINITION],
    description="A repeatable directive.",
    args={
        "service_name": GraphQLArgument(
            GraphQLNonNull(GraphQLString), description="Service Name required"
        )
    },
    is_repeatable=True,
)

# A Schema directive
ComposeDirective = CustomDirective(
    name="compose",
    locations=[DirectiveLocation.SCHEMA],
    description="A schema directive.",
    args={
        "directive_name": GraphQLArgument(
            GraphQLNonNull(GraphQLString), description="Directive Name required"
        )
    },
    is_repeatable=True,
)


@directive(target_directive=CacheDirective, max_age=100)
@directive(target_directive=AuthenticatedDirective, required=True)
class Animal(graphene.Interface):  # DirectiveLocation.INTERFACE
    age = graphene.Int(required=True)
    kind = directive(
        target_directive=CacheDirective, field=graphene.Int(required=True), max_age=60
    )  # DirectiveLocation.FIELD_DEFINITION


@directive(target_directive=CacheDirective, max_age=100)
@directive(target_directive=AuthenticatedDirective, required=True)
class TruthEnum(graphene.Enum):  # DirectiveLocation.ENUM
    A = 1
    B = 2


# Add directives to enum values [DirectiveLocation.ENUM_VALUE]
directive(field=TruthEnum.A, target_directive=AuthenticatedDirective, required=True)


@directive(target_directive=CacheDirective, max_age=100)
class Position(graphene.ObjectType):  # DirectiveLocation.OBJECT
    x = graphene.Int(required=True)
    y = directive(
        target_directive=CacheDirective, field=graphene.Int(required=True), max_age=60
    )


@directive(target_directive=CacheDirective, max_age=60)  # DirectiveLocation.OBJECT
class Human(graphene.ObjectType):
    name = graphene.String()
    born_in = graphene.String()


@directive(target_directive=CacheDirective, max_age=60)
@directive(target_directive=AuthenticatedDirective, required=True)
class HumanInput(graphene.InputObjectType):  # DirectiveLocation.INPUT_OBJECT
    born_in = graphene.String()
    name = directive(  # DirectiveLocation.INPUT_FIELD_DEFINITION
        CacheDirective,
        field=graphene.String(
            description="Test Description", deprecation_reason="Deprecated use born in"
        ),
        max_age=300,
    )


@directive(CacheDirective, max_age=200)
class Droid(graphene.ObjectType):
    name = directive(
        CacheDirective,
        field=graphene.String(
            description="Test Description", deprecation_reason="Deprecated use born in"
        ),
        max_age=300,
    )
    primary_function = graphene.String()


@directive(CacheDirective, max_age=200)
class Starship(graphene.ObjectType):
    name = graphene.String()
    length = directive(
        target_directive=CacheDirective,
        field=graphene.Int(deprecation_reason="Use another field"),
        max_age=60,
    )


@directive(target_directive=CacheDirective, max_age=500)
@directive(target_directive=AuthenticatedDirective, required=True)
class SearchResult(graphene.Union):  # DirectiveLocation.UNION
    class Meta:
        types = (Human, Droid, Starship)


@directive(target_directive=CacheDirective, max_age=500)
@directive(target_directive=AuthenticatedDirective, required=True)
class DateNewScalar(graphene.Scalar):  # DirectiveLocation.SCALAR
    pass


@directive(target_directive=KeyDirective)
@directive(target_directive=InternalDirective)
class Admin(graphene.ObjectType):
    name = graphene.String()
    password = graphene.String()


@directive(target_directive=AuthenticatedDirective, required=True)
class User(graphene.ObjectType):
    name = graphene.String()
    password = graphene.String()
    price = graphene.Field(
        graphene.String,
        currency=directive(  # DirectiveLocation.ARGUMENT_DEFINITION
            InternalDirective, field=graphene.Argument(graphene.Int)
        ),  # Argument
        country=directive(  # Multiple DirectiveLocation.ARGUMENT_DEFINITION
            target_directive=AuthenticatedDirective,
            field=directive(InternalDirective, field=graphene.Argument(graphene.Int)),
            required=True,
        ),
    )


@directive(target_directive=RepeatableDirective, service_name="ProductService")
@directive(target_directive=RepeatableDirective, service_name="CompanyService")
@directive(target_directive=AuthenticatedDirective, required=True)
class Company(graphene.ObjectType):
    established = graphene.Int(required=True)
    name = directive(
        target_directive=RepeatableDirective,
        field=directive(
            target_directive=RepeatableDirective,
            field=graphene.String(
                required=True,
                deprecation_reason="This field is deprecated and will be removed in future",
            ),
            service_name="CompanyService Field",
        ),
        service_name="ProductService Field",
    )


class Query(graphene.ObjectType):
    position = graphene.Field(Position, deprecation_reason="unused field")


schema = build_schema(
    query=Query,
    types=(
        SearchResult,
        Animal,
        Admin,
        HumanInput,
        TruthEnum,
        DateNewScalar,
        User,
        Company,
    ),
    directives=(
        CacheDirective,
        AuthenticatedDirective,
        InternalDirective,
        KeyDirective,
        RepeatableDirective,
    ),
    schema_directives=(  # extend schema directives
        SchemaDirective(
            target_directive=ComposeDirective, arguments={"directive_name": "lowercase"}
        ),
        SchemaDirective(
            target_directive=ComposeDirective, arguments={"directive_name": "uppercase"}
        ),
        SchemaDirective(
            target_directive=ComposeDirective,
            arguments={"directive_name": "pascalcase"},
        ),
    ),
)


def generate_schema() -> None:
    with open(f"{curr_dir}/complex_uses.graphql", "w") as f:
        f.write(str(schema))


if __name__ == "__main__":
    generate_schema()



from functools import partial
from typing import Any, Callable, Collection, Dict, Optional

from graphene.utils.str_converters import to_camel_case
from graphql import GraphQLArgument, GraphQLDirective
from graphql.language import ast

from .constants import ACCEPTED_TYPES, FIELD_TYPES, LOCATION_NON_FIELD_VALIDATOR
from .constants import DirectiveLocation
from .data_models import CustomDirectiveMeta
from .exceptions import (
    DirectiveInvalidArgTypeError,
    DirectiveInvalidTypeError,
    DirectiveValidationError,
)
from .parsers import parse_argument_values
from .utils import field_attribute_name, non_field_attribute_name, set_attribute_value


def CustomDirective(  # noqa
    name: str,
    locations: Collection[DirectiveLocation],
    args: Optional[Dict[str, GraphQLArgument]] = None,
    is_repeatable: bool = False,
    description: Optional[str] = None,
    extensions: Optional[Dict[str, Any]] = None,
    ast_node: Optional[ast.DirectiveDefinitionNode] = None,
    allow_all_directive_locations: bool = False,
    add_definition_to_schema: bool = True,
    non_field_validator: Callable[[Any, dict[str, Any], Any], bool] = None,
    field_validator: Callable[[Any, Any, dict[str, Any], Any], bool] = None,
    input_transform: Callable[[dict[str, Any], Any], dict[str, Any]] = None,
) -> GraphQLDirective:
    """
    Creates a GraphQLDirective

    :param name: (GraphQLDirective param)
    :param args: (GraphQLDirective param)
    :param is_repeatable: (GraphQLDirective param)
    :param description: (GraphQLDirective param)
    :param extensions: (GraphQLDirective param)
    :param ast_node: (GraphQLDirective param)

    :param locations: list[DirectiveLocation], if need to use unsupported locations, set allow_all_directive_locations True
    :param allow_all_directive_locations: Allow other DirectiveLocation other than the ones supported by library
    :param add_definition_to_schema: If false, the @directive definition is not added to the graphql schema
    :param non_field_validator: a non field validator function
                def validator (type_: graphene type, inputs: Any, schema: Schema) -> bool,
                    if validator returns False, library raises DirectiveCustomValidationError
    :param field_validator: a field validator function
                def validator (parent_type_: graphene_type, field_type_: graphene type, inputs: Any, schema: Schema) -> bool,
                    if validator returns False, library raises DirectiveCustomValidationError

    :param input_transform: a function to transform the input arg's values before usage
                def input_transform (inputs: dict[str, Any], schema: Schema) -> dict[str, Any]

    """

    if not isinstance(allow_all_directive_locations, bool):
        raise DirectiveInvalidArgTypeError(
            f"directive @{name} allow_all_directive_locations type invalid expected bool"
        )

    if not isinstance(add_definition_to_schema, bool):
        raise DirectiveInvalidArgTypeError(
            f"directive @{name} add_definition_to_schema type invalid expected bool"
        )

    if not (isinstance(non_field_validator, Callable) or non_field_validator is None):
        raise DirectiveInvalidArgTypeError(
            f"directive @{name} validator type invalid expected Callable[[GraphQLDirective, Any], bool] "
        )

    if (
        any(not isinstance(location, DirectiveLocation) for location in locations)
        and not allow_all_directive_locations
    ):
        raise DirectiveInvalidArgTypeError(
            f"directive @{name} location type invalid expected {DirectiveLocation} from graphene_directives"
        )

    if args is not None:
        camel_cased_args = {}
        for arg_name, arg_value in args.items():
            camel_cased_args[to_camel_case(arg_name)] = arg_value
    else:
        camel_cased_args = args

    target_directive = GraphQLDirective(
        name=name,
        locations=[
            location.to_graphene_directive_location()
            if isinstance(location, DirectiveLocation)
            else location
            for location in locations
        ],
        args=camel_cased_args,
        is_repeatable=is_repeatable,
        description=description,
        extensions=extensions,
        ast_node=ast_node,
    )

    valid_types = set(target_directive.locations).intersection(ACCEPTED_TYPES)
    non_field_types = set(valid_types).difference(FIELD_TYPES)
    supports_field_types = FIELD_TYPES.intersection(valid_types) != 0
    supports_non_field_types = (
        set(LOCATION_NON_FIELD_VALIDATOR.values()).intersection(valid_types) != 0
    )

    target_directive._graphene_directive = CustomDirectiveMeta(
        allow_all_directive_locations=allow_all_directive_locations,
        add_definition_to_schema=add_definition_to_schema,
        has_no_argument=target_directive.args == {},
        valid_types=valid_types,
        non_field_types=non_field_types,
        supports_field_types=supports_field_types,
        supports_non_field_types=supports_non_field_types,
        non_field_validator=non_field_validator,
        field_validator=field_validator,
        input_transform=input_transform,
    )

    # Check if target_directive.locations have accepted types
    if (
        len(valid_types) != len(target_directive.locations)
        and not allow_all_directive_locations
    ):
        invalid_types = [
            str(i) for i in set(target_directive.locations).difference(ACCEPTED_TYPES)
        ]
        raise DirectiveValidationError(
            ", ".join(
                [
                    f"{str(target_directive)}: Directives don't support types: {invalid_types}",
                    f"allowed types: {[str(i) for i in ACCEPTED_TYPES]}",
                ]
            )
        )

    return target_directive


def directive(
    target_directive: GraphQLDirective, *, field: Optional[Any] = None, **_kwargs: Any
) -> Callable:
    """
    Decorator to use to add directive a given type of field.
    """

    if not hasattr(target_directive, "_graphene_directive"):
        raise DirectiveInvalidTypeError(target_directive)

    meta_data: CustomDirectiveMeta = getattr(target_directive, "_graphene_directive")

    # Converting inputs to camel_case

    kwargs = {to_camel_case(field): value for (field, value) in _kwargs.items()}
    directive_name = str(target_directive)

    kwargs = parse_argument_values(target_directive, kwargs)

    def decorator(type_: Any) -> Any:
        if not meta_data.supports_non_field_types:
            raise DirectiveValidationError(
                f"{directive_name} cannot be used at non field level"
            )

        if not any(
            LOCATION_NON_FIELD_VALIDATOR[valid_type](type_)
            for valid_type in meta_data.non_field_types
        ):
            raise DirectiveValidationError(
                f"{directive_name} cannot be used for {type_}, valid levels are: {[str(i) for i in meta_data.valid_types]}"
            )

        set_attribute_value(
            type_=type_,
            attribute_name=non_field_attribute_name(target_directive),
            target_directive=target_directive,
            data={} if meta_data.has_no_argument else kwargs,
        )
        return type_

    if field:
        if not meta_data.supports_field_types:
            raise DirectiveValidationError(
                f"{directive_name} cannot be used at field level"
            )

        set_attribute_value(
            type_=field,
            attribute_name=field_attribute_name(target_directive),
            target_directive=target_directive,
            data={} if meta_data.has_no_argument else kwargs,
        )

        return field

    return decorator


def directive_decorator(target_directive: GraphQLDirective) -> directive:
    """
    Build a decorator for given target directive
    :param target_directive: GraphQLDirective

    Returns partial of directive(target_directive=target_directive)
    """

    return partial(directive, target_directive=target_directive)



from pathlib import Path

import graphene
from graphql import GraphQLArgument, GraphQLInt, GraphQLNonNull, GraphQLString

from graphene_directives import (
    CustomDirective,
    DirectiveLocation,
    Schema,
    build_schema,
    directive_decorator,
)

curr_dir = Path(__file__).parent


def input_transform(inputs: dict, _schema: Schema) -> dict:
    """
    def input_transform (inputs: Any, schema: Schema) -> dict,
    """
    if inputs.get("max_age") > 200:
        inputs["swr"] = 30
    return inputs


CacheDirective = CustomDirective(
    name="cache",
    locations=[
        DirectiveLocation.FIELD_DEFINITION,
        DirectiveLocation.OBJECT,
        DirectiveLocation.UNION,
    ],
    args={
        "max_age": GraphQLArgument(
            GraphQLNonNull(GraphQLInt),
            description="Specifies the maximum age for cache in seconds.",
        ),
        "swr": GraphQLArgument(
            GraphQLInt, description="Stale-while-revalidate value in seconds. Optional."
        ),
        "scope": GraphQLArgument(
            GraphQLString, description="Scope of the cache. Optional."
        ),
    },
    description="Caching directive to control cache behavior of fields or fragments.",
    input_transform=input_transform,
)


cache = directive_decorator(target_directive=CacheDirective)


@cache(max_age=500)
class Position(graphene.ObjectType):
    x = graphene.Int(required=True)
    y = cache(field=graphene.Int(required=True), max_age=60)


@cache(max_age=60)
class Human(graphene.ObjectType):
    name = graphene.String()
    born_in = graphene.String()


@cache(max_age=200)
class Droid(graphene.ObjectType):
    name = cache(field=graphene.String(), max_age=300)
    primary_function = graphene.String()


@cache(max_age=200)
class Starship(graphene.ObjectType):
    name = graphene.String()
    length = cache(field=graphene.Int(deprecation_reason="Koo"), max_age=60)


@cache(max_age=500)
class SearchResult(graphene.Union):
    class Meta:
        types = (Human, Droid, Starship)


class Query(graphene.ObjectType):
    position = graphene.Field(Position, deprecation_reason="Koo")


schema = build_schema(query=Query, types=(SearchResult,), directives=[CacheDirective])


def test_generate_schema() -> None:
    with open(f"{curr_dir}/schema_files/test_directive_input_transform.graphql") as f:
        assert str(schema) == f.read()



from pathlib import Path
from typing import Any

import graphene
from graphql import (
    GraphQLArgument,
    GraphQLBoolean,
    GraphQLInt,
    GraphQLNonNull,
    GraphQLString,
)

from graphene_directives import (
    CustomDirective,
    DirectiveLocation,
    Schema,
    SchemaDirective,
    build_schema,
    directive,
)

curr_dir = Path(__file__).parent


def validate_non_field_input(_type: Any, inputs: dict, _schema: Schema) -> bool:
    """
    def validator (type_: graphene type, inputs: Any, schema: Schema) -> bool,
    if validator returns False, library raises DirectiveCustomValidationError
    """
    if inputs.get("max_age") > 2500:
        return False
    return True


def validate_field_input(
    _parent_type: Any, _field_type: Any, inputs: dict, _schema: Schema
) -> bool:
    """
    def validator (parent_type_: graphene_type, field_type_: graphene type, inputs: Any, schema: Schema) -> bool,
    if validator returns False, library raises DirectiveCustomValidationError
    """
    if inputs.get("max_age") > 2500:
        return False
    return True


CacheDirective = CustomDirective(
    name="cache",
    locations=[
        DirectiveLocation.OBJECT,
        DirectiveLocation.INTERFACE,
        DirectiveLocation.ENUM,
        DirectiveLocation.UNION,
        DirectiveLocation.INPUT_OBJECT,
        DirectiveLocation.FIELD_DEFINITION,
        DirectiveLocation.INPUT_FIELD_DEFINITION,
        DirectiveLocation.SCALAR,
    ],
    args={
        "max_age": GraphQLArgument(
            GraphQLNonNull(GraphQLInt),
            description="Specifies the maximum age for cache in seconds.",
        ),
        "swr": GraphQLArgument(
            GraphQLInt, description="Stale-while-revalidate value in seconds. Optional."
        ),
        "scope": GraphQLArgument(
            GraphQLString, description="Scope of the cache. Optional."
        ),
    },
    description="Caching directive to control cache behavior of fields or fragments.",
    non_field_validator=validate_non_field_input,
    field_validator=validate_field_input,
)


AuthenticatedDirective = CustomDirective(
    name="authenticated",
    locations=[
        DirectiveLocation.OBJECT,
        DirectiveLocation.INTERFACE,
        DirectiveLocation.ENUM,
        DirectiveLocation.ENUM_VALUE,
        DirectiveLocation.UNION,
        DirectiveLocation.INPUT_OBJECT,
        DirectiveLocation.FIELD_DEFINITION,
        DirectiveLocation.ARGUMENT_DEFINITION,
        DirectiveLocation.SCALAR,
    ],
    args={
        "required": GraphQLArgument(
            GraphQLNonNull(GraphQLBoolean), description="Auth required"
        )
    },
    description="Auth directive to control authorization behavior.",
)

# No argument directive
HiddenDirective = CustomDirective(
    name="hidden",
    locations=[DirectiveLocation.OBJECT, DirectiveLocation.ARGUMENT_DEFINITION],
    description="Auth directive to control authorization behavior.",
)


# No Schema  directive
LinkDirective = CustomDirective(
    name="link",
    locations=[DirectiveLocation.SCHEMA],
    description="Schema directive to link files",
    args={
        "url": GraphQLArgument(
            GraphQLNonNull(GraphQLString), description="Url required"
        )
    },
)


# A Schema directive
ComposeDirective = CustomDirective(
    name="compose",
    locations=[DirectiveLocation.SCHEMA],
    description="A schema directive.",
    args={
        "directive_name": GraphQLArgument(
            GraphQLNonNull(GraphQLString), description="Directive Name required"
        )
    },
    is_repeatable=True,
)


@directive(target_directive=CacheDirective, max_age=100)
@directive(target_directive=AuthenticatedDirective, required=True)
class Animal(graphene.Interface):
    age = graphene.Int(required=True)
    kind = directive(
        target_directive=CacheDirective, field=graphene.Int(required=True), max_age=60
    )


@directive(target_directive=CacheDirective, max_age=100)
@directive(target_directive=AuthenticatedDirective, required=True)
class TruthEnum(graphene.Enum):
    A = 1
    B = 2


# Add directives to enum values
directive(field=TruthEnum.A, target_directive=AuthenticatedDirective, required=True)


@directive(target_directive=CacheDirective, max_age=100)
class Position(graphene.ObjectType):
    x = graphene.Int(required=True)
    y = directive(
        target_directive=CacheDirective, field=graphene.Int(required=True), max_age=60
    )


@directive(target_directive=CacheDirective, max_age=60)
class Human(graphene.ObjectType):
    name = graphene.String()
    born_in = graphene.String()


@directive(target_directive=CacheDirective, max_age=60)
@directive(target_directive=AuthenticatedDirective, required=True)
class HumanInput(graphene.InputObjectType):
    born_in = graphene.String()
    name = directive(
        CacheDirective,
        field=graphene.String(
            description="Test Description", deprecation_reason="Deprecated use born in"
        ),
        max_age=300,
    )


@directive(CacheDirective, max_age=200)
class Droid(graphene.ObjectType):
    name = directive(
        CacheDirective,
        field=graphene.String(
            description="Test Description", deprecation_reason="Deprecated use born in"
        ),
        max_age=300,
    )
    primary_function = graphene.String()


@directive(CacheDirective, max_age=20)
class Starship(graphene.ObjectType):
    name = graphene.String()
    length = directive(
        target_directive=CacheDirective,
        field=graphene.Int(deprecation_reason="Koo"),
        max_age=60,
    )


@directive(target_directive=CacheDirective, max_age=500)
@directive(target_directive=AuthenticatedDirective, required=True)
class SearchResult(graphene.Union):
    class Meta:
        types = (Human, Droid, Starship)


@directive(target_directive=CacheDirective, max_age=500)
@directive(target_directive=AuthenticatedDirective, required=True)
class DateNewScalar(graphene.Scalar):
    pass


@directive(target_directive=AuthenticatedDirective, required=True)
class Admin(graphene.ObjectType):
    name = graphene.String()
    password = graphene.String()
    price = graphene.Field(
        graphene.String,
        currency=directive(
            HiddenDirective,
            field=graphene.Argument(graphene.Int, deprecation_reason="Use country"),
        ),  # Argument
        country=directive(
            target_directive=AuthenticatedDirective,
            field=directive(
                HiddenDirective,
                field=graphene.Argument(graphene.Int, description="Country"),
            ),
            required=True,
        ),  # Argument Definition (Multiple directives)
    )


@directive(target_directive=AuthenticatedDirective, required=True)
class User(graphene.ObjectType):
    name = graphene.String()
    password = graphene.String()
    camel_cased = directive(HiddenDirective, field=graphene.String())
    price = graphene.Field(
        graphene.String,
        currency=directive(
            HiddenDirective, field=graphene.Argument(graphene.Int)
        ),  # Argument
        country=directive(
            target_directive=AuthenticatedDirective,
            field=directive(HiddenDirective, field=graphene.Argument(graphene.Int)),
            required=True,
        ),  # Argument Definition (Multiple directives)
    )


class Query(graphene.ObjectType):
    position = graphene.Field(Position, deprecation_reason="Koo")


schema = build_schema(
    query=Query,
    types=(SearchResult, Animal, Admin, HumanInput, TruthEnum, DateNewScalar, User),
    directives=(CacheDirective, AuthenticatedDirective, HiddenDirective, LinkDirective),
    schema_directives=(  # extend schema directives
        SchemaDirective(
            target_directive=LinkDirective,
            arguments={"url": "https://spec.graphql.org/v1.0"},
        ),
        SchemaDirective(
            target_directive=ComposeDirective, arguments={"directive_name": "lowercase"}
        ),
        SchemaDirective(
            target_directive=ComposeDirective, arguments={"directive_name": "uppercase"}
        ),
        SchemaDirective(
            target_directive=ComposeDirective, arguments={"directive_name": "lowercase"}
        ),
    ),
)


def test_generate_schema() -> None:
    with open(f"{curr_dir}/schema_files/test_directive.graphql") as f:
        assert str(schema) == f.read()



from pathlib import Path

import graphene
from graphql import (
    GraphQLArgument,
    GraphQLBoolean,
    GraphQLInt,
    GraphQLNonNull,
    GraphQLString,
)

from graphene_directives import (
    CustomDirective,
    DirectiveLocation,
    build_schema,
    directive,
)

curr_dir = Path(__file__).parent

CacheDirective = CustomDirective(
    name="cache",
    locations=[
        DirectiveLocation.OBJECT,
        DirectiveLocation.INTERFACE,
        DirectiveLocation.ENUM,
        DirectiveLocation.UNION,
        DirectiveLocation.INPUT_OBJECT,
        DirectiveLocation.FIELD_DEFINITION,
        DirectiveLocation.INPUT_FIELD_DEFINITION,
        DirectiveLocation.SCALAR,
    ],
    args={
        "max_age": GraphQLArgument(
            GraphQLNonNull(GraphQLInt),
            description="Specifies the maximum age for cache in seconds.",
        ),
        "swr": GraphQLArgument(
            GraphQLInt, description="Stale-while-revalidate value in seconds. Optional."
        ),
        "scope": GraphQLArgument(
            GraphQLString, description="Scope of the cache. Optional."
        ),
    },
    description="Caching directive to control cache behavior of fields or fragments.",
)


AuthenticatedDirective = CustomDirective(
    name="authenticated",
    locations=[
        DirectiveLocation.OBJECT,
        DirectiveLocation.INTERFACE,
        DirectiveLocation.ENUM,
        DirectiveLocation.ENUM_VALUE,
        DirectiveLocation.UNION,
        DirectiveLocation.INPUT_OBJECT,
        DirectiveLocation.FIELD_DEFINITION,
        DirectiveLocation.ARGUMENT_DEFINITION,
        DirectiveLocation.SCALAR,
    ],
    args={
        "required": GraphQLArgument(
            GraphQLNonNull(GraphQLBoolean), description="Auth required"
        )
    },
    description="Auth directive to control authorization behavior.",
)

# No argument directive
HiddenDirective = CustomDirective(
    name="hidden",
    locations=[DirectiveLocation.OBJECT, DirectiveLocation.ARGUMENT_DEFINITION],
    description="Auth directive to control authorization behavior.",
    add_definition_to_schema=False,
)


@directive(target_directive=CacheDirective, max_age=100)
@directive(target_directive=AuthenticatedDirective, required=True)
class Animal(graphene.Interface):
    age = graphene.Int(required=True)
    kind = directive(
        target_directive=CacheDirective, field=graphene.Int(required=True), max_age=60
    )


@directive(target_directive=CacheDirective, max_age=100)
@directive(target_directive=AuthenticatedDirective, required=True)
class TruthEnum(graphene.Enum):
    A = 1
    B = 2


# Add directives to enum values
directive(field=TruthEnum.A, target_directive=AuthenticatedDirective, required=True)


@directive(target_directive=CacheDirective, max_age=100)
class Position(graphene.ObjectType):
    x = graphene.Int(required=True)
    y = directive(
        target_directive=CacheDirective, field=graphene.Int(required=True), max_age=60
    )


@directive(target_directive=CacheDirective, max_age=60)
class Human(graphene.ObjectType):
    name = graphene.String()
    born_in = graphene.String()


@directive(target_directive=CacheDirective, max_age=60)
@directive(target_directive=AuthenticatedDirective, required=True)
class HumanInput(graphene.InputObjectType):
    born_in = graphene.String()
    name = directive(
        CacheDirective,
        field=graphene.String(
            description="Test Description", deprecation_reason="Deprecated use born in"
        ),
        max_age=300,
    )


class Query(graphene.ObjectType):
    position = graphene.Field(Position, deprecation_reason="Koo")


schema = build_schema(
    query=Query,
    types=(Animal, TruthEnum, HumanInput, Human, Position),
    directives=(CacheDirective, AuthenticatedDirective, HiddenDirective),
)


def test_generate_schema() -> None:
    with open(
        f"{curr_dir}/schema_files/test_arg_add_definition_to_schema.graphql"
    ) as f:
        assert str(schema) == f.read()



from pathlib import Path

import graphene
import pytest
from graphql import GraphQLArgument, GraphQLInt, GraphQLNonNull, GraphQLString

from graphene_directives import (
    CustomDirective,
    DirectiveLocation,
    DirectiveValidationError,
    directive,
)

curr_dir = Path(__file__).parent

CacheDirective = CustomDirective(
    name="cache",
    locations=[DirectiveLocation.FIELD_DEFINITION, DirectiveLocation.OBJECT],
    args={
        "max_age": GraphQLArgument(
            GraphQLNonNull(GraphQLInt),
            description="Specifies the maximum age for cache in seconds.",
        ),
        "swr": GraphQLArgument(
            GraphQLInt, description="Stale-while-revalidate value in seconds. Optional."
        ),
        "scope": GraphQLArgument(
            GraphQLString, description="Scope of the cache. Optional."
        ),
    },
    description="Caching directive to control cache behavior of fields or fragments.",
    is_repeatable=True,
)


def test_duplicate_values() -> None:
    with pytest.raises(Exception) as e_info:

        @directive(CacheDirective, max_age=20)
        @directive(CacheDirective, max_age=20)
        class _Test(graphene.ObjectType):
            pass

    assert e_info.type == DirectiveValidationError


def test_non_duplicate_values() -> None:
    @directive(CacheDirective, max_age=20)
    @directive(CacheDirective, max_age=22)
    class _Test(graphene.ObjectType):
        pass



from pathlib import Path

import graphene
import pytest
from graphql import (
    GraphQLArgument,
    GraphQLBoolean,
    GraphQLInt,
    GraphQLNonNull,
    GraphQLString,
)

from graphene_directives import (
    CustomDirective,
    DirectiveLocation,
    DirectiveValidationError,
    build_schema,
    directive,
)

curr_dir = Path(__file__).parent


CacheDirective = CustomDirective(
    name="cache",
    locations=[
        DirectiveLocation.OBJECT,
        DirectiveLocation.INTERFACE,
        DirectiveLocation.ENUM,
        DirectiveLocation.UNION,
        DirectiveLocation.INPUT_OBJECT,
        DirectiveLocation.FIELD_DEFINITION,
        DirectiveLocation.INPUT_FIELD_DEFINITION,
        DirectiveLocation.SCALAR,
    ],
    args={
        "max_age": GraphQLArgument(
            GraphQLNonNull(GraphQLInt),
            description="Specifies the maximum age for cache in seconds.",
        ),
        "swr": GraphQLArgument(
            GraphQLInt, description="Stale-while-revalidate value in seconds. Optional."
        ),
        "scope": GraphQLArgument(
            GraphQLString, description="Scope of the cache. Optional."
        ),
    },
    description="Caching directive to control cache behavior of fields or fragments.",
    is_repeatable=True,
)


AuthenticatedDirective = CustomDirective(
    name="authenticated",
    locations=[
        DirectiveLocation.OBJECT,
        DirectiveLocation.INTERFACE,
        DirectiveLocation.ENUM,
        DirectiveLocation.ENUM_VALUE,
        DirectiveLocation.UNION,
        DirectiveLocation.INPUT_OBJECT,
        DirectiveLocation.FIELD_DEFINITION,
        DirectiveLocation.ARGUMENT_DEFINITION,
        DirectiveLocation.SCALAR,
    ],
    args={
        "required": GraphQLArgument(
            GraphQLNonNull(GraphQLBoolean), description="Auth required"
        )
    },
    description="Auth directive to control authorization behavior.",
)


@directive(target_directive=CacheDirective, max_age=100)
@directive(target_directive=CacheDirective, max_age=30)
@directive(target_directive=AuthenticatedDirective, required=True)
class Animal(graphene.Interface):
    age = graphene.Int(required=True)
    kind = directive(
        target_directive=CacheDirective,
        field=directive(
            target_directive=CacheDirective,
            field=graphene.Int(
                required=True,
                deprecation_reason="This field is deprecated and will be removed in future",
            ),
            max_age=20,
        ),
        max_age=60,
    )


class Query(graphene.ObjectType):
    animals = graphene.Field(Animal, deprecation_reason="Koo")


schema = build_schema(query=Query, directives=(CacheDirective, AuthenticatedDirective))


def test_generate_schema() -> None:
    with open(f"{curr_dir}/schema_files/test_directive_with_repeatable.graphql") as f:
        assert str(schema) == f.read()


def test_non_repeatable_on_non_field() -> None:
    with pytest.raises(Exception) as e_info:

        @directive(target_directive=AuthenticatedDirective, required=True)
        @directive(target_directive=AuthenticatedDirective, required=True)
        class _TestClass(graphene.Interface):
            age = graphene.Int(required=True)
            kind = directive(
                target_directive=CacheDirective,
                field=directive(
                    target_directive=CacheDirective,
                    field=graphene.Int(
                        required=True,
                        deprecation_reason="This field is deprecated and will be removed in future",
                    ),
                    max_age=20,
                ),
                max_age=60,
            )

    assert e_info.type == DirectiveValidationError


def test_non_repeatable_on_field() -> None:
    with pytest.raises(Exception) as e_info:

        class _TestClass(graphene.Interface):
            age = graphene.Int(required=True)
            kind = directive(
                target_directive=AuthenticatedDirective,
                field=directive(
                    target_directive=AuthenticatedDirective,
                    field=graphene.Int(
                        required=True,
                        deprecation_reason="This field is deprecated and will be removed in future",
                    ),
                    required=True,
                ),
                required=False,
            )

    assert e_info.type == DirectiveValidationError



import re
from typing import Callable, Collection, Union

import graphene
from graphene import Schema as GrapheneSchema
from graphene.types.scalars import ScalarOptions
from graphene.types.union import UnionOptions
from graphene.utils.str_converters import to_camel_case, to_snake_case
from graphql import (
    DirectiveLocation,
    GraphQLArgument,
    GraphQLDirective,
    GraphQLField,
    GraphQLInputField,
    GraphQLNamedType,
    is_enum_type,
    is_input_type,
    is_interface_type,
    is_object_type,
    is_scalar_type,
    is_union_type,
    print_schema,
)
from graphql import specified_directives
from graphql.utilities.print_schema import (
    print_args,
    print_description,
    print_directive,
    print_input_value,
)

from .data_models.schema_directive import SchemaDirective
from .directive import CustomDirectiveMeta
from .exceptions import DirectiveCustomValidationError, DirectiveValidationError
from .parsers import (
    arg_camel_case,
    arg_snake_case,
    decorator_string,
    entity_type_to_fields_string,
    enum_type_to_fields_string,
    extend_schema_string,
    input_type_to_fields_string,
)
from .utils import (
    get_field_attribute_value,
    get_non_field_attribute_value,
    get_single_field_type,
    has_field_attribute,
    has_non_field_attribute,
)


class Schema(GrapheneSchema):
    def __init__(
        self,
        query: graphene.ObjectType = None,
        mutation: graphene.ObjectType = None,
        subscription: graphene.ObjectType = None,
        types: list[graphene.ObjectType] = None,
        directives: Union[Collection[GraphQLDirective], None] = None,
        auto_camelcase: bool = True,
        schema_directives: Collection[SchemaDirective] = None,
        include_graphql_spec_directives: bool = True,
    ):
        """
        Schema Definition.

        Args:
            query (Type[ObjectType]): Root query *ObjectType*. Describes entry point for fields to *read*
                data in your Schema.
            mutation (Optional[Type[ObjectType]]): Root mutation *ObjectType*. Describes entry point for
                fields to *create, update or delete* data in your API.
            subscription (Optional[Type[ObjectType]]): Root subscription *ObjectType*. Describes entry point
                for fields to receive continuous updates.
            types (Optional[Collection[Type[ObjectType]]]): List of any types to include in schema that
                may not be introspected through root types.
            directives (List[GraphQLDirective], optional): List of custom directives to include in the
                GraphQL schema.
            auto_camelcase (bool): Fieldnames will be transformed in Schema's TypeMap from snake_case
                to camelCase (preferred by GraphQL standard). Default True.
            schema_directives (Collection[SchemaDirective]): Directives that can be defined at DIRECTIVE_LOCATION.SCHEMA
                with their argument values.
            include_graphql_spec_directives (bool): Includes directives defined by GraphQL spec (@include, @skip,
                @deprecated, @specifiedBy)
        """

        self.custom_directives = directives or []
        self.schema_directives = schema_directives or []
        self.auto_camelcase = auto_camelcase
        self.directives_used: dict[str, GraphQLDirective] = {}

        directives = tuple(self.custom_directives) + (
            tuple(specified_directives) if include_graphql_spec_directives else ()
        )
        super().__init__(
            query=query,
            mutation=mutation,
            subscription=subscription,
            types=types,
            directives=directives,
            auto_camelcase=auto_camelcase,
        )

    def field_name_to_type_attribute(
        self, model: graphene.ObjectType
    ) -> Callable[[str], str]:
        """
        Create field name conversion method (from schema name to actual graphene_type attribute name).

        Args:
            model (ObjectType): model whose field name is to be converted

        Returns:
            (str) -> (str)
        """
        field_names = {}
        if self.auto_camelcase:
            field_names = {
                to_camel_case(attr_name): attr_name
                for attr_name in getattr(model._meta, "fields", [])  # noqa
            }
        return lambda schema_field_name: field_names.get(
            schema_field_name, schema_field_name
        )

    def type_attribute_to_field_name(self, attribute: str) -> str:
        """
        Create a conversion method to convert from graphene_type attribute name to the schema field name.
        """
        if self.auto_camelcase:
            return to_camel_case(attribute)
        return attribute

    def _add_argument_decorators(
        self,
        entity_name: str,
        required_directive_field_types: set[DirectiveLocation],
        args: dict[str, GraphQLArgument],
    ) -> str:
        """
        For a given field, go through all its args and see if any directive decorator needs to be added.
        """

        if not args:
            return ""

        # If every arg does not have a description, print them on one line.
        print_single_line = not any(arg.description for arg in args.values())
        indentation: str = "  "
        new_args = []

        str_field = "(" if print_single_line else "(\n"

        for i, (name, arg) in enumerate(args.items()):
            if print_single_line:
                base_str = f"{print_input_value(name, arg)} "
            else:
                base_str = (
                    print_description(arg, f"  {indentation}", not i)
                    + f"  {indentation}"
                    + f"{print_input_value(name, arg)} "
                )
            directives = []
            for directive in self.custom_directives:
                if has_field_attribute(arg, directive):
                    directive_values = get_field_attribute_value(arg, directive)
                    meta_data: CustomDirectiveMeta = getattr(
                        directive, "_graphene_directive"
                    )

                    if (
                        not required_directive_field_types.intersection(
                            set(directive.locations)
                        )
                        and len(required_directive_field_types) != 0
                    ):
                        raise DirectiveValidationError(
                            "\n".join(
                                [
                                    f"{str(directive)} cannot be used at argument {name} level",
                                    f"\tat {entity_name}",
                                    f"\tallowed: {directive.locations}",
                                    f"\trequired: {required_directive_field_types}",
                                ]
                            )
                        )

                    for directive_value in directive_values:
                        if meta_data.input_transform is not None:
                            directive_value = arg_camel_case(
                                meta_data.input_transform(
                                    arg_snake_case(directive_value), self
                                )
                            )

                        directive_str = decorator_string(directive, **directive_value)
                        directives.append(directive_str)

            new_args.append(base_str + " ".join(directives))

        if print_single_line:
            str_field += ", ".join(new_args) + ")"
        else:
            str_field += "\n".join(new_args) + f"\n{indentation})"

        return str_field

    def _add_field_decorators(self, graphene_types: set, string_schema: str) -> str:
        """
        For a given entity, go through all its fields and see if any directive decorator needs to be added.

        This method simply goes through the fields that need to be modified and replace them with their annotated
        version in the schema string representation.
        """

        for graphene_type in graphene_types:
            entity_name = graphene_type._meta.name  # noqa

            entity_type = self.graphql_schema.get_type(entity_name)
            get_field_graphene_type = self.field_name_to_type_attribute(graphene_type)

            required_directive_locations = set()

            if is_object_type(entity_type) or is_interface_type(entity_type):
                required_directive_locations.union(
                    {
                        DirectiveLocation.FIELD_DEFINITION,
                        DirectiveLocation.ARGUMENT_DEFINITION,
                    }
                )
            elif is_enum_type(entity_type):
                required_directive_locations.add(DirectiveLocation.ENUM_VALUE)
            elif is_input_type(entity_type):
                required_directive_locations.add(
                    DirectiveLocation.INPUT_FIELD_DEFINITION
                )
            else:
                continue

            if is_enum_type(entity_type):
                fields: dict = entity_type.values
            else:
                fields: dict = entity_type.fields

            str_fields = []

            for field_name, field in fields.items():
                if is_enum_type(entity_type):
                    str_field = enum_type_to_fields_string(
                        get_single_field_type(
                            entity_type, field_name, field, is_enum_type=True
                        )
                    )
                elif isinstance(field, GraphQLInputField):
                    str_field = input_type_to_fields_string(
                        get_single_field_type(entity_type, field_name, field)
                    )
                elif isinstance(field, GraphQLField):
                    str_field = entity_type_to_fields_string(
                        get_single_field_type(entity_type, field_name, field)
                    )

                    # Replace Arguments with directives
                    if hasattr(entity_type, "_fields"):
                        _arg = entity_type._fields.args[0]  # noqa
                        if hasattr(_arg, self.type_attribute_to_field_name(field_name)):
                            arg_field = getattr(
                                _arg, self.type_attribute_to_field_name(field_name)
                            )
                        else:
                            arg_field = {}

                        if (
                            hasattr(arg_field, "args")
                            and arg_field.args is not None
                            and isinstance(arg_field.args, dict)
                        ):
                            original_args = print_args(
                                args=field.args, indentation="  "
                            )
                            replacement_args = self._add_argument_decorators(
                                entity_name=entity_name,
                                required_directive_field_types=required_directive_locations,
                                args=arg_field.args,
                            )
                            str_field = str_field.replace(
                                original_args, replacement_args
                            )
                else:
                    continue

                # Check if we need to annotate the field by checking if it has the decorator attribute set on the field.
                field = getattr(
                    graphene_type, get_field_graphene_type(field_name), None
                )
                if field is None:
                    # Append the string, but skip the directives
                    str_fields.append(str_field)
                    continue

                for directive in self.custom_directives:
                    if not has_field_attribute(field, directive):
                        continue
                    directive_values = get_field_attribute_value(field, directive)

                    meta_data: CustomDirectiveMeta = getattr(
                        directive, "_graphene_directive"
                    )

                    if (
                        not required_directive_locations.intersection(
                            set(directive.locations)
                        )
                        and len(required_directive_locations) != 0
                    ):
                        raise DirectiveValidationError(
                            "\n".join(
                                [
                                    f"{str(directive)} cannot be used at field level",
                                    f"\tat {entity_name}",
                                    f"\tallowed: {directive.locations}",
                                    f"\trequired: {required_directive_locations}",
                                ]
                            )
                        )

                    for directive_value in directive_values:
                        if (
                            meta_data.field_validator is not None
                            and not meta_data.field_validator(
                                entity_type,
                                field,
                                arg_snake_case(directive_value),
                                self,
                            )
                        ):
                            raise DirectiveCustomValidationError(
                                ", ".join(
                                    [
                                        f"Custom Validation Failed for {str(directive)} with args: ({directive_value})"
                                        f"at field level {entity_name}:{field}"
                                    ]
                                )
                            )

                        if meta_data.input_transform is not None:
                            directive_value = arg_camel_case(
                                meta_data.input_transform(
                                    arg_snake_case(directive_value), self
                                )
                            )

                        str_field += (
                            f" {decorator_string(directive, **directive_value)}"
                        )

                str_fields.append(str_field)

            str_fields_annotated = "\n".join(str_fields)

            # Replace the original field declaration by the annotated one
            if is_object_type(entity_type):
                entity_type_name = "type"
                str_fields_original = entity_type_to_fields_string(entity_type)
            elif is_interface_type(entity_type):
                entity_type_name = "interface"
                str_fields_original = entity_type_to_fields_string(entity_type)
            elif is_enum_type(entity_type):
                entity_type_name = "enum"
                str_fields_original = enum_type_to_fields_string(entity_type)
            elif is_input_type(entity_type):
                entity_type_name = "input"
                str_fields_original = input_type_to_fields_string(entity_type)
            else:
                continue

            pattern = re.compile(
                r"(%s\s%s\s[^\{]*)\{\s*%s\s*\}"  # noqa
                % (entity_type_name, entity_name, re.escape(str_fields_original))
            )
            string_schema = pattern.sub(
                r"\g<1> {\n%s\n}" % str_fields_annotated, string_schema
            )
        return string_schema

    def add_non_field_decorators(
        self, non_fields_type: set[GraphQLNamedType], string_schema: str
    ) -> str:
        for non_field in non_fields_type:
            entity_name = non_field._meta.name  # noqa
            entity_type = self.graphql_schema.get_type(entity_name)

            required_directive_locations = set()

            if is_scalar_type(entity_type):
                non_field_pattern = rf"(scalar {entity_name})"
                required_directive_locations.add(DirectiveLocation.SCALAR)
            elif is_union_type(entity_type):
                non_field_pattern = rf"(union {entity_name} )"
                required_directive_locations.add(DirectiveLocation.UNION)
            elif is_object_type(entity_type):
                non_field_pattern = rf"(type {entity_name} [^\{{]*)"
                required_directive_locations.add(DirectiveLocation.OBJECT)
            elif is_interface_type(entity_type):
                non_field_pattern = rf"(interface {entity_name} [^\{{]*)"
                required_directive_locations.add(DirectiveLocation.INTERFACE)
            elif is_enum_type(entity_type):
                non_field_pattern = rf"(enum {entity_name} [^\{{]*)"
                required_directive_locations.add(DirectiveLocation.ENUM)
            elif is_input_type(entity_type):
                non_field_pattern = rf"(input {entity_name} [^\{{]*)"
                required_directive_locations.add(DirectiveLocation.INPUT_OBJECT)
            else:
                continue

            directive_annotations = []
            for directive in self.custom_directives:
                if has_non_field_attribute(non_field, directive):
                    meta_data: CustomDirectiveMeta = getattr(
                        directive, "_graphene_directive"
                    )
                    directive_values = get_non_field_attribute_value(
                        non_field, directive
                    )

                    if (
                        not required_directive_locations.intersection(
                            set(directive.locations)
                        )
                        and len(required_directive_locations) != 0
                    ):
                        raise DirectiveValidationError(
                            "\n".join(
                                [
                                    f"{str(directive)} cannot be used at non field level",
                                    f"\tat {entity_name}",
                                    f"\tallowed: {directive.locations}",
                                    f"\trequired: {required_directive_locations}",
                                ]
                            )
                        )

                    for directive_value in directive_values:
                        if (
                            meta_data.non_field_validator is not None
                            and not meta_data.non_field_validator(
                                non_field, arg_snake_case(directive_value), self
                            )
                        ):
                            raise DirectiveCustomValidationError(
                                ", ".join(
                                    [
                                        f"Custom Validation Failed for {str(directive)} with args: ({directive_value})"
                                        f"at non-field level {entity_name}"
                                    ]
                                )
                            )
                        if meta_data.input_transform is not None:
                            directive_value = arg_camel_case(
                                meta_data.input_transform(
                                    arg_snake_case(directive_value), self
                                )
                            )

                        directive_annotations.append(
                            f"{decorator_string(directive, **directive_value)}"
                        )

            annotation = " ".join(directive_annotations)
            annotation = (
                f" {annotation}" if is_scalar_type(entity_type) else f"{annotation} "
            )
            replace_str = rf"\1{annotation}"
            pattern = re.compile(non_field_pattern)
            string_schema = pattern.sub(replace_str, string_schema)

        return string_schema

    def _get_directive_applied_non_field_types(self) -> set:
        """
        Find all the directive applied non-field types from the schema.
        """
        directives_types = set()
        schema_types = {
            **self.graphql_schema.type_map,
            **{
                "Query": self.graphql_schema.query_type,
                "Mutation": self.graphql_schema.mutation_type,
            },
        }

        for schema_type in schema_types.values():
            if not hasattr(schema_type, "graphene_type"):
                continue
            for directive in self.custom_directives:
                if has_non_field_attribute(schema_type.graphene_type, directive):
                    self.directives_used[directive.name] = directive
                    directives_types.add(schema_type.graphene_type)
        return directives_types

    def _get_directive_applied_field_types(self) -> set:
        """
        Find all the directive applied field types from the schema.
        """
        directives_fields = set()
        schema_types = {
            **self.graphql_schema.type_map,
            **{
                "Query": self.graphql_schema.query_type,  # noqa
                "Mutation": self.graphql_schema.mutation_type,  # noqa
            },
        }

        for _, entity_type in schema_types.items():
            if (
                not hasattr(entity_type, "graphene_type")  # noqa:SIM101
                or isinstance(entity_type.graphene_type._meta, UnionOptions)  # noqa
                or isinstance(entity_type.graphene_type._meta, ScalarOptions)  # noqa
            ):
                continue

            fields = (
                list(entity_type.values.values())  # Enum class fields
                if is_enum_type(entity_type)
                else list(entity_type.fields)  # noqa
            )

            for field in fields:
                field_type = (
                    # auto-camelcasing can cause problems
                    getattr(entity_type.graphene_type, to_camel_case(field), None)
                    or getattr(entity_type.graphene_type, to_snake_case(field), None)
                    if not is_enum_type(entity_type)
                    else field.value
                )
                for directive_ in self.custom_directives:
                    if has_field_attribute(field_type, directive_):
                        self.directives_used[directive_.name] = directive_
                        directives_fields.add(entity_type.graphene_type)

                    # Handle Argument Decorators
                    if (
                        hasattr(field_type, "args")
                        and field_type.args is not None
                        and isinstance(field_type.args, dict)
                    ):
                        for arg_name, arg_type in field_type.args.items():
                            if has_field_attribute(arg_type, directive_):
                                if (
                                    DirectiveLocation.ARGUMENT_DEFINITION
                                    not in directive_.locations
                                ):
                                    raise DirectiveValidationError(
                                        f"{directive_} cannot be used at argument level at {entity_type}->{field}"
                                    )
                                self.directives_used[directive_.name] = directive_
                                directives_fields.add(entity_type.graphene_type)

        return directives_fields

    def get_directives_used(self) -> list[GraphQLDirective]:
        """
        Returns a list of directives used in the schema

        """
        self._get_directive_applied_field_types()
        self._get_directive_applied_non_field_types()
        return list(self.directives_used.values())

    def __str__(self):
        string_schema = ""
        string_schema += extend_schema_string(string_schema, self.schema_directives)
        string_schema += print_schema(self.graphql_schema)

        field_types = self._get_directive_applied_field_types()
        non_field_types = self._get_directive_applied_non_field_types()

        string_schema = self._add_field_decorators(field_types, string_schema)
        string_schema = self.add_non_field_decorators(non_field_types, string_schema)

        for directive in self.custom_directives:
            meta_data: CustomDirectiveMeta = getattr(directive, "_graphene_directive")
            if not meta_data.add_definition_to_schema:
                string_schema = string_schema.replace(
                    print_directive(directive) + "\n\n", ""
                )

        return string_schema.strip()



from graphql import GraphQLDirective


class DirectiveValidationError(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class DirectiveCustomValidationError(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class DirectiveInvalidTypeError(Exception):
    def __init__(self, directive: GraphQLDirective):
        message = f"Directive {str(directive)} must be build from CustomDirective(...)"
        super().__init__(message)


class DirectiveInvalidArgTypeError(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class DirectiveInvalidArgValueTypeError(Exception):
    def __init__(self, errors: list[str]):
        super().__init__("\n".join(errors))



from pathlib import Path

import graphene
import pytest
from graphql import GraphQLArgument, GraphQLInt, GraphQLNonNull, GraphQLString

from graphene_directives import CustomDirective, DirectiveLocation, directive
from graphene_directives.exceptions import DirectiveInvalidArgValueTypeError

curr_dir = Path(__file__).parent

CacheDirective = CustomDirective(
    name="cache",
    locations=[DirectiveLocation.OBJECT, DirectiveLocation.FIELD_DEFINITION],
    args={
        "max_age": GraphQLArgument(
            GraphQLNonNull(GraphQLInt),
            description="Specifies the maximum age for cache in seconds.",
        ),
        "swr": GraphQLArgument(
            GraphQLInt, description="Stale-while-revalidate value in seconds. Optional."
        ),
        "scope": GraphQLArgument(
            GraphQLString, description="Scope of the cache. Optional."
        ),
    },
    description="Caching directive to control cache behavior of fields or fragments.",
    is_repeatable=True,
)

DbCacheDirective = CustomDirective(
    name="db_cache",
    locations=[DirectiveLocation.OBJECT, DirectiveLocation.FIELD_DEFINITION],
    args={
        "max_age": GraphQLArgument(
            GraphQLNonNull(GraphQLInt),
            description="Specifies the maximum age for cache in seconds.",
            default_value=12,
        ),
        "swr": GraphQLArgument(
            GraphQLInt, description="Stale-while-revalidate value in seconds. Optional."
        ),
        "scope": GraphQLArgument(
            GraphQLString, description="Scope of the cache. Optional."
        ),
    },
    description="Caching directive to control cache behavior of fields or fragments.",
)


def test_input_argument_on_class() -> None:
    with pytest.raises(Exception) as e_info:

        @directive(target_directive=CacheDirective, required=True)
        class _TestClass(graphene.ObjectType):
            age = graphene.Int(required=True)

    assert e_info.type == DirectiveInvalidArgValueTypeError


def test_input_argument_on_field() -> None:
    with pytest.raises(Exception) as e_info:

        class _TestClass(graphene.ObjectType):
            age = graphene.Int(required=True)
            kind = directive(
                target_directive=CacheDirective,
                field=graphene.Int(
                    required=True,
                    deprecation_reason="This field is deprecated and will be removed in future",
                ),
            )

    assert e_info.type == DirectiveInvalidArgValueTypeError


def test_input_default_argument_on_class() -> None:
    @directive(target_directive=DbCacheDirective, required=True)
    class _TestClass(graphene.ObjectType):
        age = graphene.Int(required=True)


def test_input_default_argument_on_field() -> None:
    class _TestClass(graphene.ObjectType):
        age = graphene.Int(required=True)
        kind = directive(
            target_directive=DbCacheDirective,
            field=graphene.Int(
                required=True,
                deprecation_reason="This field is deprecated and will be removed in future",
            ),
        )



from typing import Collection, Type, Union

import graphene
from graphene import Schema as GrapheneSchema
from graphql import GraphQLDirective
from graphql import specified_directives

from . import DirectiveValidationError
from .data_models import SchemaDirective
from .schema import Schema


def build_schema(
    query: Union[graphene.ObjectType, Type[graphene.ObjectType]] = None,
    mutation: Union[graphene.ObjectType, Type[graphene.ObjectType]] = None,
    subscription: Union[graphene.ObjectType, Type[graphene.ObjectType]] = None,
    types: Collection[Union[graphene.ObjectType, Type[graphene.ObjectType]]] = None,
    directives: Union[Collection[GraphQLDirective], None] = None,
    auto_camelcase: bool = True,
    schema_directives: Collection[SchemaDirective] = None,
    include_graphql_spec_directives: bool = True,
) -> GrapheneSchema:
    """
    Build Schema.

    Args:
        query (Type[ObjectType]): Root query *ObjectType*. Describes entry point for fields to *read*
            data in your Schema.
        mutation (Optional[Type[ObjectType]]): Root mutation *ObjectType*. Describes entry point for
            fields to *create, update or delete* data in your API.
        subscription (Optional[Type[ObjectType]]): Root subscription *ObjectType*. Describes entry point
            for fields to receive continuous updates.
        types (Optional[Collection[Type[ObjectType]]]): List of any types to include in schema that
            may not be introspected through root types.
        directives (List[GraphQLDirective], optional): List of custom directives to include in the
            GraphQL schema.
        auto_camelcase (bool): Fieldnames will be transformed in Schema's TypeMap from snake_case
            to camelCase (preferred by GraphQL standard). Default True.
        schema_directives (Collection[SchemaDirective]): Directives that can be defined at DIRECTIVE_LOCATION.SCHEMA
            with their argument values.
        include_graphql_spec_directives (bool): Includes directives defined by GraphQL spec (@include, @skip,
            @deprecated, @specifiedBy)
    """

    _schema_directive_set: set[str] = set()
    for schema_directive in schema_directives or []:
        if schema_directive.target_directive.name in _schema_directive_set:
            if not schema_directive.target_directive.is_repeatable:
                raise DirectiveValidationError(
                    f"{schema_directive.target_directive} is not repeatable on schema"
                )
        else:
            _schema_directive_set.add(schema_directive.target_directive.name)

    directives = list(directives) if directives is not None else []

    _directive_set: set[str] = set()
    for directive in directives:
        if directive.name in _directive_set:
            raise DirectiveValidationError(f"Duplicate {directive} found")
        _directive_set.add(directive.name)

    # Validate if custom directive conflicts with graphql spec default directives
    _duplicate_default_directives = _directive_set.intersection(
        {directive.name for directive in specified_directives}
    )

    if _duplicate_default_directives:
        formatted_directive_str = [f"@{str(i)}" for i in _duplicate_default_directives]
        raise DirectiveValidationError(
            f"{formatted_directive_str} are reserved directives for client queries."
        )

    return Schema(
        query=query,
        mutation=mutation,
        subscription=subscription,
        types=types,
        directives=directives,
        auto_camelcase=auto_camelcase,
        include_graphql_spec_directives=include_graphql_spec_directives,
        schema_directives=schema_directives,
    )



import json
import re
from typing import Any, Collection, Dict, Union, cast

from graphene.utils.str_converters import to_camel_case, to_snake_case
from graphql import (
    GraphQLDirective,
    GraphQLEnumType,
    GraphQLError,
    GraphQLInputObjectType,
    GraphQLInputType,
    GraphQLInterfaceType,
    GraphQLObjectType,
    is_non_null_type,
    value_from_ast,
)
from graphql.pyutils import inspect, print_path_list
from graphql.utilities import coerce_input_value
from graphql.utilities.print_schema import (
    print_description,
    print_enum,
    print_fields,
    print_input_object,
)

from .data_models import SchemaDirective
from .exceptions import DirectiveInvalidArgValueTypeError


def _remove_block(str_fields: str) -> str:
    # Remove blocks added by `print_block`
    block_match = re.match(r" \{\n(?P<field_str>.*)\n}", str_fields, flags=re.DOTALL)
    if block_match:
        str_fields = block_match.groups()[0]
    return str_fields


def entity_type_to_fields_string(
    field: Union[GraphQLObjectType, GraphQLInterfaceType],
) -> str:
    return _remove_block(print_fields(field))


def enum_type_to_fields_string(graphene_type: GraphQLEnumType) -> str:
    fields = print_enum(graphene_type).replace(
        print_description(graphene_type) + f"enum {graphene_type.name}", ""
    )
    return _remove_block(fields)


def input_type_to_fields_string(graphene_type: GraphQLInputObjectType) -> str:
    fields = print_input_object(graphene_type).replace(
        print_description(graphene_type) + f"input {graphene_type.name}", ""
    )
    return _remove_block(fields)


def decorator_string(directive: GraphQLDirective, **kwargs: dict) -> str:
    directive_name = str(directive)
    if len(directive.args) == 0:
        return directive_name

    # Format each keyword argument as a string, considering its type
    formatted_args = [
        (
            f"{to_camel_case(key)}: "
            + (f'"{value}"' if isinstance(value, str) else json.dumps(value))
        )
        for key, value in kwargs.items()
        if value is not None and to_camel_case(key) in directive.args
    ]

    # Construct the directive string
    return f"{directive_name}({', '.join(formatted_args)})"


def extend_schema_string(
    string_schema: str, schema_directives: Collection[SchemaDirective]
) -> str:
    schema_directives_strings = []
    for schema_directive in schema_directives:
        args = parse_argument_values(
            schema_directive.target_directive,
            {
                to_camel_case(field): value
                for (field, value) in schema_directive.arguments.items()
            },
        )
        schema_directives_strings.append(
            "\t" + decorator_string(schema_directive.target_directive, **args)
        )

    if len(schema_directives_strings) != 0:
        string_schema += (
            "extend schema\n" + "\n".join(schema_directives_strings) + "\n\n"
        )

    return string_schema


def parse_argument_values(
    directive: GraphQLDirective, inputs: Dict[str, Any]
) -> dict[str, Any]:
    coerced_values: Dict[str, Any] = {}
    errors = []

    for var_name, var_arg_type in directive.args.items():
        var_type = var_arg_type.type
        snake_cased_var_name = to_snake_case(var_name)

        var_type = cast(GraphQLInputType, var_type)
        if var_name not in inputs:
            if var_arg_type.default_value:
                coerced_values[var_name] = value_from_ast(
                    var_arg_type.default_value, var_type
                )
            elif is_non_null_type(var_type):
                var_type_str = inspect(var_type)
                errors.append(
                    f"Variable '{snake_cased_var_name}' of required type '{var_type_str}'"
                    " was not provided."
                )
            continue

        value = inputs[var_name]
        if value is None and is_non_null_type(var_type):
            var_type_str = inspect(var_type)
            errors.append(
                f"Variable '{snake_cased_var_name}' of non-null type '{var_type_str}'"
                " must not be null."
            )
            continue

        def on_input_value_error(
            path: list[Union[str, int]], invalid_value: Any, error: GraphQLError
        ) -> None:
            invalid_str = inspect(invalid_value)
            prefix = (
                f"Variable '{snake_cased_var_name}' got invalid value {invalid_str}"
            )
            if path:
                prefix += f" at '{snake_cased_var_name}{print_path_list(path)}'"
            errors.append(prefix + "; " + error.message)

        coerced_values[var_name] = coerce_input_value(
            value, var_type, on_input_value_error
        )

    if errors:
        raise DirectiveInvalidArgValueTypeError(errors=errors)

    return coerced_values


def arg_camel_case(inputs: dict) -> dict:
    return {to_camel_case(k): v for k, v in inputs.items()}


def arg_snake_case(inputs: dict) -> dict:
    return {to_snake_case(k): v for k, v in inputs.items()}



from pathlib import Path

import graphene
import pytest
from graphql import GraphQLArgument, GraphQLInt, GraphQLNonNull, GraphQLString

from graphene_directives import (
    CustomDirective,
    DirectiveLocation,
    DirectiveValidationError,
    build_schema,
    directive,
)

curr_dir = Path(__file__).parent


def test_pass_reserved_directive() -> None:
    with pytest.raises(DirectiveValidationError) as e_info:
        SkipDirective = CustomDirective(
            name="skip",
            locations=[
                DirectiveLocation.OBJECT,
                DirectiveLocation.INTERFACE,
                DirectiveLocation.ENUM,
                DirectiveLocation.UNION,
                DirectiveLocation.INPUT_OBJECT,
                DirectiveLocation.FIELD_DEFINITION,
                DirectiveLocation.INPUT_FIELD_DEFINITION,
                DirectiveLocation.SCALAR,
            ],
            args={
                "max_age": GraphQLArgument(
                    GraphQLNonNull(GraphQLInt),
                    description="Specifies the maximum age for cache in seconds.",
                ),
                "swr": GraphQLArgument(
                    GraphQLInt,
                    description="Stale-while-revalidate value in seconds. Optional.",
                ),
                "scope": GraphQLArgument(
                    GraphQLString, description="Scope of the cache. Optional."
                ),
            },
            description="Caching directive to control cache behavior of fields or fragments.",
        )

        @directive(target_directive=SkipDirective, max_age=500)
        class _TestClass(graphene.ObjectType):
            age = graphene.Int(required=True)
            kind = directive(
                target_directive=SkipDirective,
                field=graphene.Int(
                    required=True,
                    deprecation_reason="This field is deprecated and will be removed in future",
                ),
                max_age=500,
            )

        _ = build_schema(types=(_TestClass,), directives=(SkipDirective,))

    assert e_info.type == DirectiveValidationError


def test_pass_reserved_directive_with_disable() -> None:
    with pytest.raises(DirectiveValidationError) as e_info:
        SkipDirective = CustomDirective(
            name="skip",
            locations=[
                DirectiveLocation.OBJECT,
                DirectiveLocation.INTERFACE,
                DirectiveLocation.ENUM,
                DirectiveLocation.UNION,
                DirectiveLocation.INPUT_OBJECT,
                DirectiveLocation.FIELD_DEFINITION,
                DirectiveLocation.INPUT_FIELD_DEFINITION,
                DirectiveLocation.SCALAR,
            ],
            args={
                "max_age": GraphQLArgument(
                    GraphQLNonNull(GraphQLInt),
                    description="Specifies the maximum age for cache in seconds.",
                ),
                "swr": GraphQLArgument(
                    GraphQLInt,
                    description="Stale-while-revalidate value in seconds. Optional.",
                ),
                "scope": GraphQLArgument(
                    GraphQLString, description="Scope of the cache. Optional."
                ),
            },
            description="Caching directive to control cache behavior of fields or fragments.",
        )

        @directive(target_directive=SkipDirective, max_age=500)
        class _TestClass(graphene.ObjectType):
            age = graphene.Int(required=True)
            kind = directive(
                target_directive=SkipDirective,
                field=graphene.Int(
                    required=True,
                    deprecation_reason="This field is deprecated and will be removed in future",
                ),
                max_age=500,
            )

        _ = build_schema(
            types=(_TestClass,),
            directives=(SkipDirective,),
            include_graphql_spec_directives=True,  # default
        )

    assert e_info.type == DirectiveValidationError


def test_pass_reserved_directive_with_disable_schema() -> None:
    CacheDirective = CustomDirective(
        name="cache",
        locations=[
            DirectiveLocation.OBJECT,
            DirectiveLocation.INTERFACE,
            DirectiveLocation.ENUM,
            DirectiveLocation.UNION,
            DirectiveLocation.INPUT_OBJECT,
            DirectiveLocation.FIELD_DEFINITION,
            DirectiveLocation.INPUT_FIELD_DEFINITION,
            DirectiveLocation.SCALAR,
        ],
        args={
            "max_age": GraphQLArgument(
                GraphQLNonNull(GraphQLInt),
                description="Specifies the maximum age for cache in seconds.",
            ),
            "swr": GraphQLArgument(
                GraphQLInt,
                description="Stale-while-revalidate value in seconds. Optional.",
            ),
            "scope": GraphQLArgument(
                GraphQLString, description="Scope of the cache. Optional."
            ),
        },
        description="Caching directive to control cache behavior of fields or fragments.",
    )

    @directive(target_directive=CacheDirective, max_age=500)
    class _TestClass(graphene.ObjectType):
        age = graphene.Int(required=True)
        kind = directive(
            target_directive=CacheDirective,
            field=graphene.Int(
                required=True,
                deprecation_reason="This field is deprecated and will be removed in future",
            ),
            max_age=500,
        )

    schema = build_schema(
        types=(_TestClass,),
        directives=(CacheDirective,),
        include_graphql_spec_directives=False,
    )

    with open(
        f"{curr_dir}/schema_files/test_directive_include_graphql_spec_directives_disable.graphql",
        "w",
    ) as f:
        f.write(str(schema))

    with open(
        f"{curr_dir}/schema_files/test_directive_include_graphql_spec_directives_disable.graphql"
    ) as f:
        assert str(schema) == f.read()


def test_pass_reserved_directive_with_enable_schema() -> None:
    CacheDirective = CustomDirective(
        name="cache",
        locations=[
            DirectiveLocation.OBJECT,
            DirectiveLocation.INTERFACE,
            DirectiveLocation.ENUM,
            DirectiveLocation.UNION,
            DirectiveLocation.INPUT_OBJECT,
            DirectiveLocation.FIELD_DEFINITION,
            DirectiveLocation.INPUT_FIELD_DEFINITION,
            DirectiveLocation.SCALAR,
        ],
        args={
            "max_age": GraphQLArgument(
                GraphQLNonNull(GraphQLInt),
                description="Specifies the maximum age for cache in seconds.",
            ),
            "swr": GraphQLArgument(
                GraphQLInt,
                description="Stale-while-revalidate value in seconds. Optional.",
            ),
            "scope": GraphQLArgument(
                GraphQLString, description="Scope of the cache. Optional."
            ),
        },
        description="Caching directive to control cache behavior of fields or fragments.",
    )

    @directive(target_directive=CacheDirective, max_age=500)
    class _TestClass(graphene.ObjectType):
        age = graphene.Int(required=True)
        kind = directive(
            target_directive=CacheDirective,
            field=graphene.Int(
                required=True,
                deprecation_reason="This field is deprecated and will be removed in future",
            ),
            max_age=500,
        )

    schema = build_schema(
        types=(_TestClass,),
        directives=(CacheDirective,),
        include_graphql_spec_directives=True,  # default
    )

    with open(
        f"{curr_dir}/schema_files/test_directive_include_graphql_spec_directives.graphql",
        "w",
    ) as f:
        f.write(str(schema))

    with open(
        f"{curr_dir}/schema_files/test_directive_include_graphql_spec_directives.graphql"
    ) as f:
        assert str(schema) == f.read()



from copy import deepcopy
from typing import Any
from typing import Union

from graphql import (
    GraphQLDirective,
    GraphQLEnumType,
    GraphQLField,
    GraphQLInputField,
    GraphQLInputObjectType,
    GraphQLObjectType,
)

from .exceptions import DirectiveValidationError


# ruff: noqa: ANN401


def get_single_field_type(
    entity: Union[GraphQLEnumType, GraphQLInputObjectType, GraphQLObjectType],
    field_name: str,
    field_type: Union[GraphQLInputField, GraphQLField],
    is_enum_type: bool = False,
) -> Union[GraphQLEnumType, GraphQLInputObjectType, GraphQLObjectType]:
    """
    Generates the schema for a type with just one given field
    """
    new_entity = deepcopy(entity)
    setattr(
        new_entity, "values" if is_enum_type else "fields", {field_name: field_type}
    )
    return new_entity


def field_attribute_name(target_directive: GraphQLDirective) -> str:
    return f"_directive_{target_directive.name}_field"


def non_field_attribute_name(target_directive: GraphQLDirective) -> str:
    return f"_directive_{target_directive.name}_non_field"


def has_field_attribute(type_: Any, target_directive: GraphQLDirective) -> bool:
    return hasattr(type_, field_attribute_name(target_directive))


def has_non_field_attribute(type_: Any, target_directive: GraphQLDirective) -> bool:
    return hasattr(type_, non_field_attribute_name(target_directive))


def set_attribute_value(
    type_: Any, attribute_name: str, target_directive: GraphQLDirective, data: dict
) -> None:
    if hasattr(type_, attribute_name):
        if not target_directive.is_repeatable:
            raise DirectiveValidationError(
                f"{target_directive} is not repeatable, at: {type_}"
            )
        kwargs_list: list = getattr(type_, attribute_name)

        for prev_data in kwargs_list:
            if prev_data == data:
                raise DirectiveValidationError(
                    f"{target_directive} is got duplicate values {data}, at: {type_}"
                )

        kwargs_list.append(data)
    else:
        setattr(type_, attribute_name, [data])


def get_field_attribute_value(
    type_: Any, target_directive: GraphQLDirective
) -> list[dict]:
    return getattr(type_, field_attribute_name(target_directive))


def get_non_field_attribute_value(
    type_: Any, target_directive: GraphQLDirective
) -> list[dict]:
    return getattr(type_, non_field_attribute_name(target_directive))



from .constants import DirectiveLocation
from .data_models import SchemaDirective
from .directive import ACCEPTED_TYPES
from .directive import CustomDirective, directive, directive_decorator
from .exceptions import DirectiveCustomValidationError, DirectiveValidationError
from .main import build_schema
from .schema import Schema

__all__ = [
    "build_schema",
    "Schema",
    "CustomDirective",
    "SchemaDirective",
    "directive_decorator",
    "directive",
    "ACCEPTED_TYPES",
    "DirectiveLocation",
    "DirectiveCustomValidationError",
    "DirectiveValidationError",
]



from pathlib import Path

import pytest
from graphql import (
    DirectiveLocation as GrapheneDirectiveLocation,
    GraphQLArgument,
    GraphQLInt,
    GraphQLNonNull,
    GraphQLString,
)

from graphene_directives import CustomDirective, DirectiveLocation
from graphene_directives.exceptions import DirectiveInvalidArgTypeError

curr_dir = Path(__file__).parent


def test_invalid_location_types() -> None:
    with pytest.raises(Exception) as e_info:
        CustomDirective(
            name="cache",
            locations=[GrapheneDirectiveLocation.FIELD, DirectiveLocation.OBJECT],
            args={
                "max_age": GraphQLArgument(
                    GraphQLNonNull(GraphQLInt),
                    description="Specifies the maximum age for cache in seconds.",
                ),
                "swr": GraphQLArgument(
                    GraphQLInt,
                    description="Stale-while-revalidate value in seconds. Optional.",
                ),
                "scope": GraphQLArgument(
                    GraphQLString, description="Scope of the cache. Optional."
                ),
            },
            description="Caching directive to control cache behavior of fields or fragments.",
        )

    assert e_info.type == DirectiveInvalidArgTypeError


def test_invalid_location_types_allow() -> None:
    CustomDirective(
        name="cache",
        locations=[GrapheneDirectiveLocation.FIELD, DirectiveLocation.OBJECT],
        args={
            "max_age": GraphQLArgument(
                GraphQLNonNull(GraphQLInt),
                description="Specifies the maximum age for cache in seconds.",
            ),
            "swr": GraphQLArgument(
                GraphQLInt,
                description="Stale-while-revalidate value in seconds. Optional.",
            ),
            "scope": GraphQLArgument(
                GraphQLString, description="Scope of the cache. Optional."
            ),
        },
        description="Caching directive to control cache behavior of fields or fragments.",
        allow_all_directive_locations=True,
    )


def test_correct_location() -> None:
    CustomDirective(
        name="cache",
        locations=list(DirectiveLocation),
        args={
            "max_age": GraphQLArgument(
                GraphQLNonNull(GraphQLInt),
                description="Specifies the maximum age for cache in seconds.",
            ),
            "swr": GraphQLArgument(
                GraphQLInt,
                description="Stale-while-revalidate value in seconds. Optional.",
            ),
            "scope": GraphQLArgument(
                GraphQLString, description="Scope of the cache. Optional."
            ),
        },
        description="Caching directive to control cache behavior of fields or fragments.",
        allow_all_directive_locations=False,
    )

    CustomDirective(
        name="cache2",
        locations=list(GrapheneDirectiveLocation),
        args={
            "max_age": GraphQLArgument(
                GraphQLNonNull(GraphQLInt),
                description="Specifies the maximum age for cache in seconds.",
            ),
            "swr": GraphQLArgument(
                GraphQLInt,
                description="Stale-while-revalidate value in seconds. Optional.",
            ),
            "scope": GraphQLArgument(
                GraphQLString, description="Scope of the cache. Optional."
            ),
        },
        description="Caching directive to control cache behavior of fields or fragments.",
        allow_all_directive_locations=True,
    )



from dataclasses import dataclass
from typing import Any

from graphql import DirectiveLocation as GrapheneDirectiveLocation
from graphql import GraphQLDirective

from ..exceptions import DirectiveValidationError


@dataclass
class SchemaDirective:
    target_directive: GraphQLDirective
    arguments: dict[str, Any]

    def __post_init__(self):
        if GrapheneDirectiveLocation.SCHEMA not in self.target_directive.locations:
            raise DirectiveValidationError(
                ". ".join(
                    [
                        f"{self.target_directive} cannot be used as schema directive",
                        "Missing DirectiveLocation.SCHEMA in locations",
                    ]
                )
            )



from nox import parametrize, session


@session(python=["3.9", "3.10", "3.11", "3.12"])
@parametrize("graphene", ("3.0", "3.1", "3.2", "3.3"))
def tests(session, graphene):  # noqa
    session.install(f"graphene=={graphene}")
    session.install("pytest", ".")
    session.run("pytest", "tests/")



from dataclasses import dataclass
from typing import Any, Callable, Union

from graphql import DirectiveLocation as GrapheneDirectiveLocation


@dataclass
class CustomDirectiveMeta:
    allow_all_directive_locations: bool
    add_definition_to_schema: bool
    has_no_argument: bool
    valid_types: set[GrapheneDirectiveLocation]
    non_field_types: set[GrapheneDirectiveLocation]
    supports_field_types: bool
    supports_non_field_types: bool
    input_transform: Union[
        Callable[[dict[str, Any], Any], dict[str, Any]], None
    ]  # (args, schema) -> args
    non_field_validator: Union[
        Callable[[Any, dict[str, Any], Any], bool], None
    ]  # (type, args, schema) -> valid
    field_validator: Union[
        Callable[[Any, Any, dict[str, Any], Any], bool], None
    ]  # (parent_type, field_type, args, schema) -> valid



from enum import Enum

import graphene
from graphql import DirectiveLocation as GrapheneDirectiveLocation

LOCATION_NON_FIELD_VALIDATOR = {
    GrapheneDirectiveLocation.SCHEMA: lambda t: issubclass(t, graphene.Schema),
    GrapheneDirectiveLocation.SCALAR: lambda t: issubclass(t, graphene.Scalar),
    GrapheneDirectiveLocation.OBJECT: lambda t: issubclass(t, graphene.ObjectType),
    GrapheneDirectiveLocation.INTERFACE: lambda t: issubclass(t, graphene.Interface),
    GrapheneDirectiveLocation.UNION: lambda t: issubclass(t, graphene.Union),
    GrapheneDirectiveLocation.ENUM: lambda t: issubclass(t, graphene.Enum),
    GrapheneDirectiveLocation.INPUT_OBJECT: lambda t: issubclass(
        t, graphene.InputObjectType
    ),
}


FIELD_TYPES = {
    GrapheneDirectiveLocation.FIELD_DEFINITION,
    GrapheneDirectiveLocation.INPUT_FIELD_DEFINITION,
    GrapheneDirectiveLocation.ENUM_VALUE,
    GrapheneDirectiveLocation.ARGUMENT_DEFINITION,
}

ACCEPTED_TYPES = FIELD_TYPES.union(set(LOCATION_NON_FIELD_VALIDATOR.keys()))

TYPE_STRING_MAPPING = {
    graphene.Scalar: "scalar",
    graphene.Union: "union",
    graphene.ObjectType: "type",
    graphene.Interface: "interface",
    graphene.Enum: "enum",
    graphene.InputObjectType: "input",
}


class DirectiveLocation(Enum):
    """The enum type representing the directive location values."""

    SCHEMA = "schema"
    SCALAR = "scalar"
    OBJECT = "object"
    FIELD_DEFINITION = "field definition"
    ARGUMENT_DEFINITION = "argument definition"
    INTERFACE = "interface"
    UNION = "union"
    ENUM = "enum"
    ENUM_VALUE = "enum value"
    INPUT_OBJECT = "input object"
    INPUT_FIELD_DEFINITION = "input field definition"

    def to_graphene_directive_location(self) -> GrapheneDirectiveLocation:
        return GrapheneDirectiveLocation(self.value)
